#!/usr/bin/python
#  Copyright 2005, PathScale, Inc. All rights reserved.
#
#  Permission is granted to PathScale customers to use this software in 
#  source and binary forms, with or without modification, provided that 
#  the following conditions are met: 
#  
#  + All copies of source code must retain the above copyright notice, 
#    this list of conditions and the following disclaimer. 
#  
#  + Customer has a valid license agreement from PathScale, Inc. for the 
#    software product with which this software is originally distributed. 
#  
#  + Redistribution of original or modified versions to third parties is 
#    prohibited without specific prior written permission of PathScale, 
#    Inc. 
#  
#  + If customer makes modifications to this software, and provides those 
#    modifications to PathScale, customer assigns to PathScale, Inc. 
#    ownership of the provided modifications and all intellectual property 
#    rights embodied in those modifications. 
#  
#  + The name of PathScale, Inc. may not be used to endorse or promote 
#    products derived from this software without specific prior written 
#    permission. 
#  
#  THIS SOFTWARE IS PROVIDED BY PATHSCALE, INC. AND ITS LICENSORS "AS IS" 
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
#  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
#  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PATHSCALE, INC. OR ITS 
#  LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
#  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
#  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
#  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
#  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
#  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

import cPickle as pickle
import errno
import os
import select
import socket
import sys

program_name = os.path.basename(sys.argv[0])
hostname = socket.gethostname()


class Socket:
    PROTO = 2

    def __init__(self, sock = None):
        self._sock = sock
        if sock is None:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._file = self._sock.makefile('r+')

    def dump(self, data):
        pickle.dump(data, self._file, self.PROTO)
        self._file.flush()

    def load(self):
        return pickle.load(self._file)

    def __getattr__(self, name):
        return getattr(self._sock, name, None) or getattr(self._file, name)

    def close(self):
        self._sock.close()
        self._file.close()


class Server:
    def __init__(self):
        self.registry = {}
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind(('0.0.0.0', 0))
        self.socket.listen(5)

    def register(self, function, name = None):
        self.registry[name or function.__name__] = function

    def get_port(self):
        return self.socket.getsockname()[1]

    def handle_request(self):
        sock = Socket(self.socket.accept()[0])
        name, args, kwargs = sock.load()
        try:
            ret = self.registry[name](*args, **kwargs)
            sock.dump((False, ret))
        except:
            sock.dump((True, sys.exc_info()[1]))
        sock.close()


class RemoteException(Exception):
    def __init__(self, exc):
        self.exc = exc

    def __str__(self):
        return '%s: %s' % (self.exc.__class__.__name__, self.exc)


class Client:
    def __init__(self, host, port):
        self._addr = (host, port)

    def __getattr__(self, name):
        class Proxy:
            def __call__(_, *args, **kwargs):
                sock = Socket()
                sock.connect(self._addr)
                sock.dump((name, args, kwargs))
                raised, result = sock.load()
                sock.close()
                if raised:
                    raise RemoteException(result)
                return result
        return Proxy()


class Fd:
    def __init__(self, fd):
        self.fd = fd

    def fileno(self):
        return self.fd


class Selector:
    def interrupted(self):
        pass

    def select_loop(self, socket, readable, writable, mapping):
        while self.loop_test():
            try:
                r, w, e = select.select(readable, [], [])
                for fp in r:
                    if fp is socket:
                        fd, data = socket.load()
                        if fd is None:
                            self.control_message(data)
                        elif data:
                            os.write(writable[fd].fileno(), data)
                        else:
                            os.close(writable[fd].fileno())
                    else:
                        data = os.read(fp.fileno(), 4096)
                        socket.dump((mapping[fp.fileno()], data))
                        if not data:
                            readable.remove(fp)
                            os.close(fp.fileno())
            except select.error, err:
                if err.args[0] != errno.EINTR:
                    raise
                self.interrupted()


def proxy(addr = ''):
    host, port = (addr or os.getenv('MPD_ADDRESS')).split(':')
    return Client(host, int(port))


if __name__ == '__main__':
    print proxy().ping()
