/*
	Copyright (C) 2006. QLogic Corporation.  All rights reserved.
	Permission is granted to QLogic customers to use this software in 
	source and binary forms, with or without modification, provided that 
	the following conditions are met: 
	
	+ All copies of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer. 
	
	+ Customer has a valid license agreement from QLogic Corporation. for the 
	  software product with which this software is originally distributed. 
	
	+ Redistribution of original or modified versions to third parties is 
	  prohibited without specific prior written permission of QLogic
	  Corporation. 
	
	+ If customer makes modifications to this software, and provides those 
	  modifications to QLogic, customer assigns to QLogic Corporation. 
	  ownership of the provided modifications and all intellectual property 
	  rights embodied in those modifications. 
	
	+ The name of QLogic Corporation. may not be used to endorse or promote 
	  products derived from this software without specific prior written 
	  permission. 
	
	THIS SOFTWARE IS PROVIDED BY QLOGIC CORPORATION. AND ITS LICENSORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL QLOGIC CORPORATION. OR ITS 
	LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
	NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/


#include <sys/types.h>
#include <sys/time.h>
#include <inttypes.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include <assert.h>

#include "psm.h"
#include "psm_mq.h"

#define SEC_IN_NS   1000000000ULL
#define MAX_BUFSIZE 4*1024*1024

static char r_buf[MAX_BUFSIZE] __attribute__((aligned(64)));
static char s_buf[MAX_BUFSIZE] __attribute__((aligned(64)));

uint64_t get_stamp_us(void);
uint64_t elapsed_us(uint64_t t_start);

void do_latency_test(psm_mq_t mq, psm_epaddr_t epaddr_me, 
		psm_epaddr_t epaddr_peer, int am_sender,
		int msgsize, int iters);


/*
 * This is where you add code to connect to an out-of-band connection management
 * utility and get a new uuid.
 * This program will not work until that functionality is implemented
*/
int your_connection_manager_init(psm_uuid_t uuid)
{
    fprintf(stderr, "You need to supply a connection manager call\n");
    return -1;
}

/*
 * Push the new endpoint id to our central connection manager and obtain the
 * list of endpoint IDs.
 * This program will not work until that functionality is implemented
 */
int inform_your_connection_manager(psm_epid_t epid, psm_epid_t **epidlist)
{
    fprintf(stderr, "You need to supply a connection manager notification call\n");
    return -1;
}

int
main(int argc, char *argv[])
{
    int verno_minor = PSM_VERNO_MINOR;
    int verno_major = PSM_VERNO_MAJOR;
    int numeps;
    psm_epid_t myepid, *epids;
    psm_error_t *errors;
    int i, am_sender = 0;
    int msgsize = 0;
    int iters = 50000;
    uint64_t timeout = 20 * SEC_IN_NS;
    psm_epaddr_t *epaddrs;
    psm_uuid_t uuid;
    psm_epaddr_t epaddr_me = NULL, epaddr_peer = NULL;

    psm_ep_t ep;
    psm_mq_t mq;

    int opt;

    while ((opt = getopt(argc, argv, "+i:m:b")) != -1) {
	switch(opt) {
	    case 'i':
		iters = (int) strtoul(optarg, NULL, 0);
		break;
	    case 'm':
		msgsize = (int) strtoul(optarg, NULL, 0);
		if (msgsize > MAX_BUFSIZE) {
		    fprintf(stderr, 
			"Message size must be between 0 and %d bytes\n",
			MAX_BUFSIZE);
		    exit(1);
		}
		break;
	    default:
		fprintf(stderr, "Usage: %s [-i iters] [-m msgsize]\n", argv[0]);
		exit(1);
		break;
	}
    }

    psm_init(&verno_major, &verno_minor);

    /* 
     * This is where you add code to connect to an out-of-band connection management
     * utility and get a new uuid.
     *
     * The UUID should be randomly generated from a single source and
     * distributed to all endpoints that wish to participate in a common job.
     *
     * WARNING WARNING: Implementors are expected to write their own mechanisms
     * for generating uuids and distributing epids to all endpoints in a job.
     *
     */
    if (your_connection_manager_init(uuid) != 0)
	return -1;

    /*
     * Open a new InfiniBand endpoint handle and identifing endpoint ID.
     * We initialize the endpoint with a uuid and use default parameters in
     * opening the endpoint.
     */
    psm_ep_open(uuid, NULL, &ep, &myepid);

    /*
     * Push the new endpoint id to our central connection manager and obtain the
     * list of endpoint IDs.
     */
    numeps = inform_your_connection_manager(myepid, &epids);
    if (numeps < 1) {
	fprintf(stderr, "Couldn't get epids from MPIRUN\n");
	return -1;
    }
    else if (numeps < 2) {
	fprintf(stderr, "Need at least 2 endpoints to run this test\n");
	return -1;
    }
    else if (numeps % 2 != 0) {
	fprintf(stderr, "Need an even number of endpoints to run this test\n");
	return -1;
    }

    /*
     * We initialize the Matched Queues interface.  We don't have any specific
     * options to set and we want total ordering on all packets since we're not
     * providing a hint as to where the MPI context bits are in the tag.
     *
     * Can happen before or after connect, as long is it happens before any of
     * the MQ component functions.
     */
    psm_mq_init(ep, PSM_MQ_ORDERMASK_ALL, NULL, 0, &mq);

    /* Prepare to connect a list of endpoint addresses and call connect. We
     * don't look at the return value here since we rely on PSM's internal
     * error handler (it will kill the job if something goes wrong).
     *
     * Here we connect all the endpoints since that's what a typical MPI
     * library would do.  For this test, we could have only connected to our
     * peer if we would have wanted to.  Note that unlike Mellanox InfiniBand
     * RC, InfiniPath connects are connectionless and hold a few hundred bytes
     * of state.  There's no need to do any dynamic connection management.
     */
    epaddrs = (psm_epaddr_t *) malloc(sizeof(psm_epaddr_t) * numeps);
    errors = (psm_error_t *) malloc(sizeof(psm_error_t) * numeps);
    psm_ep_connect(ep, numeps, epids, NULL, errors, epaddrs, timeout);
    free(errors); /* don't do any error checking.  Eventually, version 1.02
		   * will allow users to pass NULL if they don't care about the
		   * errors here. Most users will probably care about the
		   * errors, though */

    /*
     * Find my rank in the list of endpoint IDs (this assumes that the mpirun
     * spawning utility maintains a equivalent ordered set on all endpoints).
     *
     * Sender is in the first half, receiver is in the second half.
     */
    for (i = 0; i < numeps; i++)  {
	if (epids[i] == myepid) {
	    epaddr_me = epaddrs[i];
	    if (i < numeps/2) {
		am_sender = 1;
		epaddr_peer = epaddrs[i + numeps/2];
	    }
	    else {
		am_sender = 0;
		epaddr_peer = epaddrs[i - numeps/2];
	    }
	    break;
	}
    }
    if (i == numeps) {
	fprintf(stderr, "Couldn't find myself in epid list\n");
	return -1;
    }

    do_latency_test(mq, epaddr_me, epaddr_peer, am_sender, msgsize, iters);

    psm_mq_finalize(mq);
    psm_ep_close(ep, PSM_EP_CLOSE_GRACEFUL, 5 * SEC_IN_NS);
    psm_finalize();
    
    return 0;
}

uint64_t 
get_stamp_us(void) 
{
    uint64_t retval; 
    struct timeval tv;
    gettimeofday(&tv, NULL);
    retval = ((uint64_t)tv.tv_sec) * 1000000 + (uint64_t)tv.tv_usec;
    return retval;
}

uint64_t
elapsed_us(uint64_t t_start)
{
    return (get_stamp_us() - t_start);
}

#define MQ_TAG	       1ULL
#define MQ_TAGSEL_ALL  0ULL
#define MQ_FLAGS_NONE  0
#define MQ_NO_CONTEXT_PTR   ((void *)NULL)

void 
do_latency_test(psm_mq_t mq, psm_epaddr_t epaddr_me, 
		psm_epaddr_t epaddr_peer, int am_sender,
		int msgsize, int iters)
{
    uint64_t t_start, t_total;
    int i;
    int warmup_iters = 1000;
    int am_sender_iter = am_sender;
    int iters_total;
    psm_mq_req_t    req;

    /* touch the data */
    memset(r_buf, 0, msgsize);
    memset(s_buf, 0, msgsize);

    am_sender_iter = am_sender;
    warmup_iters *= 2;
    iters_total = iters * 2 + warmup_iters;

    t_start = get_stamp_us();
    for (i = 0; i < iters_total; i++) {
	if (i == warmup_iters)
	    t_start = get_stamp_us();

	/* 
	 * PSM has a send where the data can be overwritten once the send
	 * returns ("blocking" in MPI parlance)
	 */
	if (am_sender_iter)
	    psm_mq_send(mq, epaddr_peer, 
			MQ_FLAGS_NONE,  /* no flags, not a sync send */
		        MQ_TAG, /* don't care tag */
			s_buf, msgsize);
	else {
	    /*
	     * PSM does not have any blocking receive, 
	     */
	    psm_mq_irecv(mq, 
		    MQ_TAG, /* don't care tag */
		    MQ_TAGSEL_ALL, /* always successfully tag match */
		    MQ_FLAGS_NONE, /* no flags */
		    r_buf,
		    msgsize,
		    MQ_NO_CONTEXT_PTR,
		    &req);
	    psm_mq_wait(&req, NULL); /* don't care about status */
	}
	am_sender_iter = 1 - am_sender_iter;
    }

    t_total = elapsed_us(t_start);

    if (am_sender)
	printf("Latency size=%8d bytes, time=%4.4f us\n",
	    msgsize, (double) t_total / iters / 2);

    return;
}
		   
